// Copyright 2006 all rights reserved; see LICENSE file for BSD-style license

package edu.berkeley.sbp;
import edu.berkeley.sbp.*;
import edu.berkeley.sbp.util.*;
import edu.berkeley.sbp.Parser.Table.*;
import edu.berkeley.sbp.Sequence.Position;
import java.io.*;
import java.util.*;
import java.lang.reflect.*;

/** implements Tomita's Graph Structured Stack */
class GSS {

    Input input;
    public GSS(Input input) { this.input = input; }
    public Input getInput() { return input; }

    // FIXME: right now, these are the performance bottleneck
    HashMapBag<Integer,Sequence>       performed       = new HashMapBag<Integer,Sequence>();

    /** FIXME */
    Forest.Many finalResult;

    /** corresponds to a positions <i>between tokens</i> the input stream; same as Tomita's U_i's */
    class Phase<Tok> implements Invokable<State, Result, Object>, IntegerMappable, GraphViz.ToGraphViz, Iterable<Node> {

        public ArrayList<Reduction> reductionQueue = new ArrayList<Reduction>();

        public void invoke(State st, Result result, Object o) {
            //shifts++;
            good |= next.newNode(result, st, false);
        }

        /** the token immediately after this phase */
        final Tok token;
        final int pos;

        public IntPairMap<Node> hash;  /* ALLOC */
        private boolean good;
        private Phase next = null;
        private Phase prev;
        private Input.Location location;
        private Input.Location nextLocation;
        private Input.Location prevLocation;
        
        private Forest forest;

        public Phase(Phase prev, Phase previous, Tok token, Input.Location location,
                     Input.Location nextLocation, Forest forest) throws ParseFailed {
            this.prevLocation = prev==null ? location : prev.getLocation();
            this.prev = prev;
            this.forest = forest;
            this.pos = previous==null ? 0 : previous.pos+1;
            this.token = token;
            this.location = location;
            this.nextLocation = nextLocation;
            performed.clear();
            hash = new IntPairMap<Node>();
            good = false;
            finalResult = null;
            if (prev != null) prev.shift(this, forest);
        }

        public boolean isFrontier() { return next==null; }
        public boolean isDone() throws ParseFailed {
            if (token != null) return false;
            if (token==null && finalResult==null)
                ParseFailed.error("unexpected end of file", this);
            return true;
        }

        public Input.Location getPrevLocation() { return prevLocation; }
        public Input.Location getLocation() { return location; }
        public Input.Region   getRegion() { return getPrevLocation().createRegion(getLocation()); }
        public Input.Location getNextLocation() { return nextLocation; }

        /** perform all shift operations, adding promoted nodes to <tt>next</tt> */
        public void shift(Phase next, Forest result) throws ParseFailed {
            // this massively improves GC performance
            if (prev!=null) prev.hash = null;
            this.next = next;
            for(Node n : hash.values()) {
                if (token == null && n.state().isAccepting()) {
                    if (finalResult==null) finalResult = new Forest.Many();
                    for(Result r : n)
                        finalResult.merge(r.getForest());
                }
                if (token == null) continue;
                n.state().invokeShifts(token, this, new Result(result, n, null), null);
            }
            if (!good && token!=null) ParseFailed.error("unexpected character", this);
            if (token==null && finalResult==null) ParseFailed.error("unexpected end of file", this);
        }

        /** perform all reduction operations */
        public void reduce() throws ParseFailed {
            Reduction last = null;
            while(!reductionQueue.isEmpty()) {
                Reduction r = null;

                // ugly
                OUTER: for(int i=0; i<reductionQueue.size(); i++) {
                    for(int j=0; j<reductionQueue.size(); j++) {
                        if (i==j) continue;
                        if (reductionQueue.get(i).compareTo(reductionQueue.get(j)) > 0)
                            continue OUTER;
                    }
                    r = reductionQueue.get(i);
                    reductionQueue.remove(r);
                    break;
                }

                /*
                if (last == null) last = r;
                else if (r.compareTo(last) > 0) last = r;
                else if (r.compareTo(last) < 0) {
                    if (r.targetPhase() != null)
                        System.out.println("err " + last.compareTo(r) + " " + last.targetPhase().pos() +
                                           " "    + r.targetPhase().pos() + " " + pos);
                }
                */

                r.perform();
            }
        }

        public void newNodeFromReduction(Result result, State state, boolean fromEmptyReduction, Position reduction) {
            int pos = result.phase().pos;
            Sequence owner = reduction.owner();
            for(Sequence s : owner.hates)
                if (performed.contains(pos, s))
                    return;
            for(Sequence s : owner.needs)
                if (!performed.contains(pos, s))
                    return;
            if (owner.needed_or_hated && !performed.contains(pos, owner))
                performed.add(pos, owner);
            if (state!=null)
                newNode(result, state, fromEmptyReduction);
        }

        /** add a new node (merging with existing nodes if possible)
         *  @param parent             the parent of the new node
         *  @param result             the SPPF result corresponding to the new node
         *  @param state              the state that the new node is in
         *  @param fromEmptyReduction true iff this node is being created as a result of a reduction of length zero (see GRMLR paper)
         *  @param start              the earliest part of the input contributing to this node (used to make merging decisions)
         */
        public boolean newNode(Result result, State state, boolean fromEmptyReduction) {
            Node p = hash.get(state, result.phase());
            if (p != null) { p.merge(result); return true; }
            do {
                if (token != null && state.canShift(token)) break;
                if (state.isAccepting()) break;
                if (token==null) break;
                if (!state.canReduce(token)) return false;
            } while(false);
            new Node(Phase.this, result, state, fromEmptyReduction);  // ALLOC
            return true;
        }

        public int toInt() { return pos+1; }
        public int size() { return hash==null ? 0 : hash.size(); }
        public int pos() { return pos; }
        public Tok getToken() { return token; }
        public Iterator<Node> iterator() { return hash.iterator(); }
        public GSS getGSS() { return GSS.this; }

        // GraphViz //////////////////////////////////////////////////////////////////////////////

        public GraphViz.Node toGraphViz(GraphViz gv) {
            if (gv.hasNode(this)) return gv.createNode(this);
            GraphViz.Group g = gv.createGroup(this);
            g.label = "Phase " + pos;
            g.color = "gray";
            g.cluster = true;
            return g;
        }
        public boolean isTransparent() { return false; }
        public boolean isHidden() { return false; }

    }

}
